/********************************************************************
BeClock KWin - effect plugin

Copyright (C) 2009,2010 Thomas Lübking <thomas.luebking@web.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*********************************************************************/

#include "beclock.h"
#include "dbus.h"

#include <kwinconfig.h>
#include <kwinglutils.h>
#include <kwinxrenderutils.h>

#include <kaction.h>
#include <kactioncollection.h>
#include <kconfiggroup.h>
#include <kglobal.h>
#include <klocale.h>
#include <kstandarddirs.h>
#include <ksystemtimezone.h>

#include <QApplication>
#include <QDesktopWidget>
#include <QDBusConnection>
#include <QPainter>
#include <QRadialGradient>
#include <QTimer>
#include <QTimerEvent>
#include <QVector4D>

#include <cmath>

#include <kdebug.h>

static const int animTime = 300;

using namespace KWin;

KWIN_EFFECT( beclock, BeClock )

BeClock::BeClock() :
  imActive( false )
, iHideOnHover( false )
, iPollTheMouse( false )
, myFadeTimer( 0 )
, myAlpha( 1.0 )
, myLastUpdateTime( -1 )
, myCountDown( 0 )
{
    QDBusConnection::sessionBus().registerObject("/BeClock", this);
    new BeClockAdaptor(this);
    myBufferXR[0] = myBufferXR[1] = 0;
    myBufferGL[0] = myBufferGL[1] = 0;
    KActionCollection* actionCollection = new KActionCollection( this );
    KAction* a = static_cast< KAction* >( actionCollection->addAction( "BeClock" ));
    a->setText( i18n("BeClock" ));
    a->setGlobalShortcut( KShortcut( Qt::CTRL + Qt::SHIFT + Qt::Key_F12 ));
    connect( a, SIGNAL( triggered(bool) ), this, SLOT( toggle() ) );
    connect( QApplication::desktop(), SIGNAL(resized(int)), this, SLOT(updateGeometry()) );

    QMetaObject::invokeMethod(this, "init", Qt::QueuedConnection);
}

void
BeClock::init()
{
    reconfigure( ReconfigureAll );
}

BeClock::~BeClock()
{
    delete myBufferXR[0]; delete myBufferXR[1];
    delete myBufferGL[0]; delete myBufferGL[1];
    foreach( ElectricBorder border, myElectricBorders )
        effects->unreserveElectricBorder( border );
}

static bool synthetic = false;

bool
BeClock::borderActivated( ElectricBorder border )
{
    if ( !synthetic && (iPollTheMouse || myRemainingAnimationTime > 0) )
        return myElectricBorders.contains( border ); // not twice...

    if ( myElectricBorders.contains( border ) )
    {
        myCurrentCorner = border;
        myGeometry = cornerRect( border );
        myTargetGeometry = myGeometry;
        switch ( border )
        {
            case ElectricTopLeft:
                myTargetGeometry.moveTopLeft( myGeometry.bottomRight() ); break;
            case ElectricTopRight:
                myTargetGeometry.moveTopRight( myGeometry.bottomLeft() ); break;
            case ElectricBottomLeft:
                myTargetGeometry.moveBottomLeft( myGeometry.topRight() ); break;
            case ElectricBottomRight:
                myTargetGeometry.moveBottomRight( myGeometry.topLeft() ); break;
            case ElectricTop:
                myTargetGeometry.moveTop( myGeometry.bottom() ); break;
            case ElectricRight:
                myTargetGeometry.moveRight( myGeometry.left() ); break;
            case ElectricLeft:
                myTargetGeometry.moveLeft( myGeometry.right() ); break;
            case ElectricBottom:
                myTargetGeometry.moveBottom( myGeometry.top() ); break;
            default:
                break;
        }
        myRemainingAnimationTime = animTime;
        if (!synthetic)
        {
            pollTheMouse(true);
            if ( !imActive )
                show(false);
        }
        return true;
    }
    return false;
}

void BeClock::countdown(uint secs)
{
    if (!imActive && !myElectricBorders.isEmpty())
    {
        show(false);
        synthetic = true;
        borderActivated(myElectricBorders.at(0));
        synthetic = false;
    }
    if ( secs > 23*3600+59*60+59 )
        secs = 23*3600+59*60+59;
    myLastUpdateTime = -1;
    myCountDown = secs;
}

int BeClock::screen() const
{
    if (myScreen < 0 || myScreen > effects->numScreens())
        return effects->activeScreen();
    return myScreen;
}

void
BeClock::updateGeometry()
{

    const QRect desktop = effects->clientArea( ScreenArea, screen(), effects->currentDesktop() );

    if ( myAlign & (Qt::AlignHCenter|Qt::AlignVCenter) )
        myGeometry.moveCenter( desktop.center() );

    if ( myAlign & Qt::AlignRight )
        myGeometry.moveRight( desktop.right() );
    else if ( myAlign & Qt::AlignLeft )
        myGeometry.moveLeft( desktop.left() );

    if ( myAlign & Qt::AlignTop )
        myGeometry.moveTop( desktop.top() );
    else if ( myAlign & Qt::AlignBottom )
        myGeometry.moveBottom( desktop.bottom() );

    myGeometry.translate( myOffset.x, myOffset.y );
}

QRect
BeClock::cornerRect( ElectricBorder border )
{
    const QRect desktop = effects->clientArea( ScreenArea, screen(), effects->currentDesktop() );
    QRect rect = myGeometry;
    switch ( border )
    {
        case ElectricTopLeft:
            rect.moveBottomRight( desktop.topLeft() ); break;
        case ElectricTopRight:
            rect.moveBottomLeft( desktop.topRight() ); break;
        case ElectricBottomLeft:
            rect.moveTopRight( desktop.bottomLeft() ); break;
        case ElectricBottomRight:
            rect.moveTopLeft( desktop.bottomRight() ); break;
        case ElectricTop:
            rect.moveBottom( desktop.top() ); break;
        case ElectricRight:
            rect.moveLeft( desktop.right() ); break;
        case ElectricLeft:
            rect.moveRight( desktop.left() ); break;
        case ElectricBottom:
            rect.moveTop( desktop.bottom() ); break;
        default: break;
    }
    return rect;
}

static void
drawOutlinedArc( QPainter &p, const QRect &r, int sta, int swa, int w, const QColor color, const QColor outline )
{
    QPainter::CompositionMode mode = p.compositionMode();
    p.setPen( QPen(outline, w, Qt::SolidLine, Qt::RoundCap) );
    p.drawArc( r, sta, swa );

    w -= 1;
    p.setCompositionMode( QPainter::CompositionMode_DestinationOut );
    p.setPen( QPen(Qt::black, w, Qt::SolidLine, Qt::RoundCap) );
    p.drawArc( r, sta, swa );

    p.setCompositionMode( mode );
    p.setPen( QPen(color, w, Qt::SolidLine, Qt::RoundCap) );
    p.drawArc( r, sta, swa );
}


#if 0
QPainterPath path;
QFont fnt;
fnt.setBold(true);
fnt.setPointSize( fnt.pointSize()*r.width()/QFontMetrics(fnt).width(apm) );
QFontMetrics fm(fnt);
QRect tr = fm.boundingRect( apm );

path.addText( r.x() + (r.width()-tr.width())/2,
r.y() + (r.height() - tr.height())/2 + fm.ascent(),
fnt, apm );

bc.setAlpha( 64 );
c.setAlpha( alpha/6 );
p.setPen( bc );
p.setBrush( c );
p.drawPath( path );

path = QPainterPath();
fnt = QFont();
fnt.setBold(false);
fnt.setPointSize( fnt.pointSize()*r.width()/QFontMetrics(fnt).width(time) );
fm = QFontMetrics(fnt);
tr = fm.boundingRect( apm );
path.addText( r.x() + (r.width()-tr.width())/2,
r.y() + (r.height() - tr.height())/2 + fm.ascent(),
fnt, time );

bc.setAlpha( 128 );
c.setAlpha( alpha*2/3 );
p.setPen( bc );
p.setBrush( c );
p.drawPath( path );
#endif


void
BeClock::attend()
{
    if (myRemainingAttention)
    {
        if (imActive)
            hide();
        else
            show();
        QTimer::singleShot(myAttentionTimeout[imActive], this, SLOT(attend()));
        --myRemainingAttention;
    }
    else
    {
        if ( imActive && !myElectricBorders.isEmpty() )
            iPollTheMouse = mouseIsInAciveCorner( QCursor::pos() );
    }
}

void
BeClock::getAttention(int n)
{
    myRemainingAttention = 2*n;
    if (!imActive && !myElectricBorders.isEmpty())
    {
        synthetic = true;
        borderActivated(myElectricBorders.at(0));
        synthetic = false;
    }
    attend();
}

void
BeClock::mouseChanged(const QPoint &pos,
                      const QPoint&, Qt::MouseButtons, Qt::MouseButtons, Qt::KeyboardModifiers, Qt::KeyboardModifiers)
{
    if (!iPollTheMouse)
        return;
    if ( iHideOnHover )
    {
        float lastAlpha = myAlpha;
        if ( myGeometry.contains( pos ) )
        {
            QPoint c = myGeometry.center();
            const float d = sqrt( pow(c.x()-pos.x(), 2) + pow(c.y()-pos.y(), 2) );
            const float r = myGeometry.height()/2.0;
            myAlpha = qMin(1.0, pow(1.0-(r-d)/r, 6) );
        }
        else
            myAlpha = 1.0;
        if ( lastAlpha != myAlpha )
            effects->addRepaint( myGeometry );
    }
    else if ( !mouseIsInAciveCorner( pos ) )
    {
        pollTheMouse(false);
        myTargetGeometry = cornerRect( myCurrentCorner );
        myRemainingAnimationTime = animTime;
        if ( imActive ) // maybe turned off otherwise
            QTimer::singleShot( animTime, this, SLOT(plainHide()) );
        effects->addRepaintFull();
    }
}

bool
BeClock::mouseIsInAciveCorner( const QPoint &pos )
{
    if ( myElectricBorders.isEmpty() )
        return false;

    const int d = 25;
    const QRect desktop = effects->clientArea( ScreenArea, screen(), effects->currentDesktop() );

    if ( pos.x() < desktop.x() + d && pos.y() < desktop.y() + d )
        return myElectricBorders.contains( ElectricTopLeft );
    if ( pos.x() > desktop.right() - d && pos.y() < desktop.y() + d )
        return myElectricBorders.contains( ElectricTopRight );
    if ( pos.x() < desktop.x() + d && pos.y() > desktop.bottom() - d )
        return myElectricBorders.contains( ElectricBottomLeft );
    if ( pos.x() > desktop.right() - d && pos.y() > desktop.bottom() - d )
        return myElectricBorders.contains( ElectricBottomRight );

    return false;
}
void
BeClock::pollTheMouse(bool b)

{
    if (iPollTheMouse == b)
        return;

    iPollTheMouse = b;
    if (iPollTheMouse)
    {
#if KDE_IS_VERSION(4,6,30)
        connect    ( effects, SIGNAL(mouseChanged(const QPoint&, const QPoint&, Qt::MouseButtons, Qt::MouseButtons,
                                     Qt::KeyboardModifiers, Qt::KeyboardModifiers)),
                                SLOT(mouseChanged(const QPoint&, const QPoint&, Qt::MouseButtons, Qt::MouseButtons,
                                     Qt::KeyboardModifiers, Qt::KeyboardModifiers)) );
#endif
        effects->startMousePolling(); // start doing expensive shit
    }
    else
    {
#if KDE_IS_VERSION(4,6,30)
        disconnect ( effects, SIGNAL(mouseChanged(const QPoint&, const QPoint&, Qt::MouseButtons, Qt::MouseButtons,
                                     Qt::KeyboardModifiers, Qt::KeyboardModifiers)),
                     this,      SLOT(mouseChanged(const QPoint&, const QPoint&, Qt::MouseButtons, Qt::MouseButtons,
                                     Qt::KeyboardModifiers, Qt::KeyboardModifiers)) );
#endif
        effects->stopMousePolling(); // stop doing expensive shit
    }
}

void
BeClock::reconfigure( ReconfigureFlags )
{
    pollTheMouse(false);
    myRemainingAnimationTime = 0;

    KConfigGroup conf = effects->effectConfig("BeClock");

    foreach( ElectricBorder border, myElectricBorders )
        effects->unreserveElectricBorder( border );
    myElectricBorders.clear();

    QList<int> borderList;
    borderList = conf.readEntry( "ElectricBorders", borderList );
    for ( int i = 0; i < borderList.size(); ++i )
    {
        ElectricBorder border = (ElectricBorder)borderList.at(i);
        myElectricBorders << border;
        effects->reserveElectricBorder( border );
    }

    iHideOnHover = myElectricBorders.isEmpty() && conf.readEntry( "HideOnHover", false );
    // TODO the rings are naturally 12h, so should be the digit default, yesno?
    iUse24h = conf.readEntry( "24h", /*false*/ !KGlobal::locale()->use12Clock() );
    // TODO though this is the only sanne date format, should one still prefer the preset as default?
    myDatePattern = conf.readEntry( "DatePattern", "yyy-MM-dd" );
    QDateTime cdt = QDateTime::currentDateTime();
    QString homeZone = conf.readEntry( "HomeZone", KSystemTimeZones::local().name() );
    QDateTime hcdt = KSystemTimeZones::local().convert(KSystemTimeZones::zone(homeZone), cdt);
    myHomeTimeDiff = cdt.secsTo(hcdt) / 60;
    myFont.fromString(conf.readEntry("Font", myFont.toString()));
    iShowTheDate = conf.readEntry( "ShowDate", false );
    iShowDigits = conf.readEntry( "ShowDigits", true );
    iShowRings = conf.readEntry( "ShowRings", true );
    iShowTheSeconds = conf.readEntry( "ShowSeconds", false );
    iChime = conf.readEntry( "Chime", false );
    myForeground = conf.readEntry( "Foreground", QColor(0,0,0,255) );
    myBackground = conf.readEntry( "Background", QColor(255,255,255,255) );
    myAlign = conf.readEntry("Align", (int)(Qt::AlignRight|Qt::AlignTop));
    myScreen = conf.readEntry( "Screen", -1 );
    iShowTheDay = conf.readEntry( "ShowDayOfWeek", true );
    myCountDownAttention = conf.readEntry( "CountDownAttention", 3 );
    const uint attentionPeriod = qMax(1500, conf.readEntry( "AttentionPeriod", 3600 ));
    myAttentionTimeout[0] = 2*attentionPeriod/3;
    myAttentionTimeout[1] = attentionPeriod/3;
    const int s = conf.readEntry("Size", 96);
    myGeometry = QRect(0,0,s,s);
    myType = (Type)conf.readEntry( "Type", (int)Rings );

    myOffset.x = conf.readEntry("X", 0);
    myOffset.y = conf.readEntry("Y", 0);
    if ( myAlign & Qt::AlignRight )
        myOffset.x = -myOffset.x;
    if ( myAlign & Qt::AlignBottom )
        myOffset.y = -myOffset.y;
    updateGeometry();

#if 0
    if ( effects->compositingType() == XRenderCompositing )
    {
        delete myBufferXR[0];
        myBufferXR[0] = new QPixmap(s,s);
        myBufferXR[0]->fill( Qt::transparent );
    }
#endif

    const int d = qMax( 2, myGeometry.width()/8 );
    QImage *img = new QImage(QSize(d,d), QImage::Format_ARGB32);
    mySecondGeometry = img->rect();
    img->fill( Qt::transparent );
    QPainter p(img);

    QColor c = myForeground;
    const int alpha = c.alpha();
    c.setAlpha( alpha/4 );
    QColor bc = myBackground;
    bc.setAlpha( alpha/6 );
    const int penw = qMax(1,d/8);
    p.setPen( QPen(bc, penw) );
    p.setBrush( c );
    p.setRenderHint( QPainter::Antialiasing );
    p.drawEllipse( img->rect().adjusted(penw, penw, -penw, -penw) );
    p.end();

    if ( effects->compositingType() == OpenGLCompositing )
    {
        if ( myBufferGL[1] )
            myBufferGL[1]->load( *img );
        else
            myBufferGL[1] = new GLTexture( *img );
    }

    if ( effects->compositingType() == XRenderCompositing )
    {
        delete myBufferXR[1];
        myBufferXR[1] = new QPixmap( QPixmap::fromImage( *img ) );
    }

    delete img;

    updateBuffer();

    if ( myElectricBorders.isEmpty() xor imActive )
        toggle(); // activate by default if not mousedriven
    if ( iHideOnHover )
        pollTheMouse(true);

    effects->addRepaintFull();
}
#ifndef KWIN_HAVE_OPENGLES
static float vanilla_matrix[16];
#endif

void BeClock::prePaintScreen( ScreenPrePaintData& data, int time )
{
    myPaintCounter = 0;
    // everyone else may go first, so we can check for transformations and save matrix storage on simple screens
    effects->prePaintScreen( data, time );
    if ( imActive )
    {
        if ( myRemainingAnimationTime > 0 )
        {
            myRemainingAnimationTime -= time;
            if ( myRemainingAnimationTime < 0 )
                myRemainingAnimationTime = 0;
            const int t = animTime;
            const int t1 = animTime - myRemainingAnimationTime;
            const int t2 = myRemainingAnimationTime;
            myGeometry.moveTo( (t1 * myTargetGeometry.x() + t2 * myGeometry.x()) / t,
                               (t1 * myTargetGeometry.y() + t2 * myGeometry.y()) / t );
        }
#ifndef KWIN_HAVE_OPENGLES
        if ( effects->compositingType() == OpenGLCompositing && ( data.mask & PAINT_SCREEN_TRANSFORMED ) )
            glGetFloatv(GL_MODELVIEW_MATRIX, vanilla_matrix);
#endif
    }
}

static inline
void xRender( QPixmap *pix, const QRect &r, float a )
{
    if (a < 1.0)
        XRenderComposite( display(), PictOpOver,
                          pix->x11PictureHandle(), xRenderBlendPicture(a), effects->xrenderBufferPicture(),
                          0, 0, 0, 0, r.x(), r.y(), r.width(), r.height() );
    else
        XRenderComposite( display(), PictOpOver,
                          pix->x11PictureHandle(), 0, effects->xrenderBufferPicture(),
                          0, 0, 0, 0, r.x(), r.y(), r.width(), r.height() );
}

void BeClock::paintScreen( int mask, QRegion region, ScreenPaintData& data )
{
    effects->paintScreen( mask, region, data ); // paint normal screen

#if KDE_IS_VERSION(4,7,30)
    if ( !(imActive) )
#else
    if ( !(imActive && region.intersects(myGeometry)) )
#endif
        return;

#if KDE_IS_VERSION(4,6,30)
    if (++myPaintCounter > 14)
#else
    if (++myPaintCounter > 1)
#endif
        return;

    // and ourself above
    if ( effects->compositingType() == OpenGLCompositing && myBufferGL[0] )
    {
#ifndef KWIN_HAVE_OPENGLES
        if ( mask & PAINT_SCREEN_TRANSFORMED )
        {
            glPushMatrix();
            glLoadMatrixf(vanilla_matrix);
        }
        glPushAttrib( GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT );
#endif
        glEnable( GL_BLEND );
#if KDE_IS_VERSION(4,6,30)
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
#else
        glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
#endif
        ShaderManager *sm = ShaderManager::instance();
        GLShader *shader;
        if (sm->isValid() && (shader = sm->getBoundShader()))
            shader->setUniform(GLShader::ModulationConstant, QVector4D(myAlpha, myAlpha, myAlpha, myAlpha));
        GLint mode;
        if (!shader && myAlpha < 1.0)
        {
#ifndef KWIN_HAVE_OPENGLES
            glGetTexEnviv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &mode);
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
#if KDE_IS_VERSION(4,6,30)
            glColor4f(myAlpha, myAlpha, myAlpha, myAlpha);
#else
            glColor4f(1.0, 1.0, 1.0, myAlpha);
#endif
#endif
        }

        if (iShowTheSeconds && !myCountDown)
        {
            myBufferGL[1]->bind();
            myBufferGL[1]->render( infiniteRegion(), mySecondGeometry );
            myBufferGL[1]->unbind();
        }

        myBufferGL[0]->bind();
        myBufferGL[0]->render( infiniteRegion(), myGeometry );
        myBufferGL[0]->unbind();
#ifndef KWIN_HAVE_OPENGLES
        if (!shader && myAlpha < 1.0)
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, mode);
        glPopAttrib();

        if ( mask & PAINT_SCREEN_TRANSFORMED )
            glPopMatrix();
#endif
    }

    if ( effects->compositingType() == XRenderCompositing && myBufferXR[0] )
    {
        if (iShowTheSeconds && !myCountDown)
            xRender( myBufferXR[1], mySecondGeometry, myAlpha );
        xRender( myBufferXR[0], myGeometry, myAlpha );
    }
}

void
BeClock::postPaintScreen()
{
    if ( imActive && myRemainingAnimationTime > 0 )
        effects->addRepaintFull();
    effects->postPaintScreen();
}

void
BeClock::timerEvent( QTimerEvent *te )
{

    if ( te->timerId() == myFadeTimer )
    {
        bool fadeDone = false;
        if ( iFadeIn )
        {
            myAlpha += 0.1;
            fadeDone = myAlpha >= 1.0;
        }
        else
        {
            myAlpha -= 0.1;
            fadeDone = myAlpha <= 0.0;
        }
        if (fadeDone)
        {
            killTimer( myFadeTimer );
            myFadeTimer = 0;
            if (!iFadeIn)
            {
                killTimer( myClock );
                myClock = 0;
                imActive = false;
            }
        }
        effects->addRepaint( myGeometry );
        return;
    }


    if ( te->timerId() != myClock )
        return;

    const int m = myCountDown > 23*60+59 ? myCountDown/60 : (myCountDown ? myCountDown : QTime::currentTime().minute());
    bool update = iShowTheSeconds;
    if ( iShowTheSeconds )
    {
        float a = -2*M_PI*(QTime::currentTime().second()-15)/60;
        int r = (myGeometry.width()-(mySecondGeometry.width() + mySecondGeometry.width()/3)) / 2;
        mySecondGeometry.moveCenter( myGeometry.center() + QPoint(qRound(r * cos(a)), -qRound(r * sin(a))) );
    }
    if ( m != myLastUpdateTime )
    {
        update = true;
        updateBuffer();
        myLastUpdateTime = m;
    }
    if ( myCountDown )
    {
        --myCountDown;
        if (!myCountDown)
        {
            myLastUpdateTime = -1;
            if (!(iPollTheMouse || myElectricBorders.isEmpty())) {
                hide(true);
                imActive = false;
            }
            getAttention(myCountDownAttention);
        }
    }
    else if (m == 0 && iChime && !QTime::currentTime().second())
        getAttention(20);
    if (update)
        effects->addRepaint( myGeometry );
}

void
BeClock::hide(bool smooth)
{
    iFadeIn = false;
    if (smooth)
    {
        if ( !myFadeTimer )
            myFadeTimer = startTimer( 40 );
    }
    else
    {
        imActive = false;
        myRemainingAnimationTime = 0;
        killTimer( myClock );
        myClock = 0;
    }
    effects->addRepaint( myGeometry );
}

void
BeClock::show(bool smooth)
{
    imActive = true;
    myLastUpdateTime = -1;
    myClock = startTimer( 1000 );
    QTimerEvent te( myClock );
    timerEvent(&te);
    iFadeIn = true;
    if (smooth)
    {
        myAlpha = 0.0;
        if ( !myFadeTimer )
            myFadeTimer = startTimer( 40 );
    }
    else
        myAlpha = 1.0;
    effects->addRepaint( myGeometry );
}

void
BeClock::toggle()
{
    if (imActive)
        hide();
    else
        show();
}


/*
// Exponential blur, Jani Huhtanen, 2006 ==========================
*  expblur(QImage &img, int radius)
*
*  In-place blur of image 'img' with kernel of approximate radius 'radius'.
*  Blurs with two sided exponential impulse response.
*
*  aprec = precision of alpha parameter in fixed-point format 0.aprec
*  zprec = precision of state parameters zR,zG,zB and zA in fp format 8.zprec
*/

template<int aprec, int zprec>
static inline void blurinner(unsigned char *bptr, int &zR, int &zG, int &zB, int &zA, int alpha)
{
    int R,G,B,A;
    R = *bptr;
    G = *(bptr+1);
    B = *(bptr+2);
    A = *(bptr+3);

    zR += (alpha * ((R<<zprec)-zR))>>aprec;
    zG += (alpha * ((G<<zprec)-zG))>>aprec;
    zB += (alpha * ((B<<zprec)-zB))>>aprec;
    zA += (alpha * ((A<<zprec)-zA))>>aprec;

    *bptr =     zR>>zprec;
    *(bptr+1) = zG>>zprec;
    *(bptr+2) = zB>>zprec;
    *(bptr+3) = zA>>zprec;
}

template<int aprec,int zprec>
static inline void blurrow( QImage & im, int line, int alpha)
{
    int zR,zG,zB,zA;

    QRgb *ptr = (QRgb *)im.scanLine(line);

    zR = *((unsigned char *)ptr    )<<zprec;
    zG = *((unsigned char *)ptr + 1)<<zprec;
    zB = *((unsigned char *)ptr + 2)<<zprec;
    zA = *((unsigned char *)ptr + 3)<<zprec;

    for(int index=1; index<im.width(); index++)
        blurinner<aprec,zprec>((unsigned char *)&ptr[index],zR,zG,zB,zA,alpha);

    for(int index=im.width()-2; index>=0; index--)
        blurinner<aprec,zprec>((unsigned char *)&ptr[index],zR,zG,zB,zA,alpha);
}

template<int aprec, int zprec>
static inline void blurcol( QImage & im, int col, int alpha)
{
    int zR,zG,zB,zA;

    QRgb *ptr = (QRgb *)im.bits();
    ptr+=col;

    zR = *((unsigned char *)ptr    )<<zprec;
    zG = *((unsigned char *)ptr + 1)<<zprec;
    zB = *((unsigned char *)ptr + 2)<<zprec;
    zA = *((unsigned char *)ptr + 3)<<zprec;

    for(int index=im.width(); index<(im.height()-1)*im.width(); index+=im.width())
        blurinner<aprec,zprec>((unsigned char *)&ptr[index],zR,zG,zB,zA,alpha);

    for(int index=(im.height()-2)*im.width(); index>=0; index-=im.width())
        blurinner<aprec,zprec>((unsigned char *)&ptr[index],zR,zG,zB,zA,alpha);
}

void
expblur( QImage &img, int radius )
{
    if(radius<1)
        return;

    static const int aprec = 16; static const int zprec = 7;

    // Calculate the alpha such that 90% of the kernel is within the radius. (Kernel extends to infinity)
    int alpha = (int)((1<<aprec)*(1.0f-expf(-2.3f/(radius+1.f))));

    for(int row=0;row<img.height();row++)
        blurrow<aprec,zprec>(img,row,alpha);

    for(int col=0;col<img.width();col++)
        blurcol<aprec,zprec>(img,col,alpha);
}
// ======================================================

QImage *
BeClock::halo( QImage *img, int r )
{
    QImage *halo = new QImage(*img);
    QRgb *pixel = (QRgb*)halo->bits();
    int n = halo->width() * halo->height();

    QColor bc = myBackground;
    QRgb rgba = bc.rgba();
    bc.setAlpha(0);
    QRgb trans = bc.rgba();

#define ALPHA qAlpha(pixel[i])
        for (int i = 0; i < n; ++i)
            pixel[i] = ALPHA > 96 ? rgba : trans;
#undef ALPHA
    expblur( *halo, r );
    QPainter p(halo);
    int v = qMax(bc.red(), qMax(bc.green(), bc.blue()));
    if ( v > 128 )
        p.drawImage(0,0, *halo);
    p.setCompositionMode( QPainter::CompositionMode_DestinationOut );
    p.drawImage(0,0, *img);
    p.end();
    return halo;
}

QImage *
BeClock::digitImage( const QDateTime &dt, QRect r )
{
    QFont fnt = myFont;
    QColor c = myForeground;
    const int alpha = c.alpha();
    const bool showDate = iShowTheDate && !myCountDown;

    int h = dt.time().hour();
    if ( !(iUse24h || myCountDown) )
    {   // i hope this is now correct ... anglo-america sucks! ;-P
        if (h > 12) h -= 12;
        else if (!h) h = 12;
    }


    QImage *img = new QImage(myGeometry.size(), QImage::Format_ARGB32);
    img->fill( Qt::transparent );
    QPainter p(img);

    // THE "UNDERLAY" ==============================================================================
    QString underlay;
    if ( myCountDown )
        underlay = "-T";
    else if ( iShowTheDay )
        underlay = QDate::shortDayName( dt.date().dayOfWeek() ).left(2);
    else if ( !iUse24h )
        underlay = (dt.time().hour() < 12 ? "AM" : "PM");
    if ( !underlay.isEmpty() )
    {
        fnt.setBold(true);
        const QRect tr = QFontMetrics(fnt).tightBoundingRect(underlay);
        const float f = qMin( (float)r.width()/tr.width(), (float)r.height()/tr.height() );
        fnt.setPointSize( fnt.pointSize()*f );
        p.setFont(fnt);
        c.setAlpha( alpha/6 );
        p.setPen( c );
        p.drawText(r.adjusted(-1000,-1000,1000,1000), Qt::AlignCenter, underlay );
    }

    // DATE ========================================================================================
    if ( showDate )
    {
        const QString date = dt.date().toString(myDatePattern);
        fnt = myFont;
        fnt.setBold(false);
        fnt.setPointSize( fnt.pointSize()*r.width()/QFontMetrics(fnt).width(date) );
        int dy = QFontMetrics(fnt).height();
        r.translate(0,4*dy/5);
        p.setFont(fnt);
        c.setAlpha( alpha/2 );
        p.setPen( c );
        p.drawText(r.adjusted(-1000,-1000,1000,1000), Qt::AlignCenter, date );
        r.translate(0,-5*dy/4);
    }

    // TIME ========================================================================================
    const QString time = QString("%1:%2").arg( h, 2, 10, QLatin1Char('0') ).arg( dt.time().minute(), 2, 10, QLatin1Char('0') );
    fnt = myFont;
    fnt.setBold(false);
    fnt.setPointSize( fnt.pointSize()*r.width()/QFontMetrics(fnt).width(time) );
    p.setFont(fnt);
    c.setAlpha( alpha*2/3 );
    p.setPen( c );
    p.drawText(r.adjusted(-1000,-1000,1000,1000), Qt::AlignCenter, time );

    p.end();

    QImage *hlo = halo( img, 9 );
    p.begin(hlo);
    p.drawImage(0,0, *img);
    p.end();
    delete img;
    return hlo;
}

#define SET_BITS(_C_, _EXP_) t = dt.time()._EXP_; for (int i = 0; i <= 4; ++i) matrix[_C_][i] = t & (1<<(3-i))

QImage *
BeClock::binaryImage( const QDateTime &dt )
{
    static bool matrix[4][4];
    int t;
    SET_BITS(0, hour() / 10);
    SET_BITS(1, hour() % 10);
    SET_BITS(2, minute() / 10);
    SET_BITS(3, minute() % 10);

    QImage *img = new QImage(myGeometry.size(), QImage::Format_ARGB32);
    img->fill( Qt::transparent );
    QRect r(img->rect()); r.adjust(2,2,-2,-2);
    int d = qMin(r.width(), r.height()) / 5 - 2;
    QPainterPath dot; dot.addEllipse(0,0,d,d);
    d += (qMin(r.width(), r.height()) - 5*d) / 3;

    QPainter p( img );
    p.setRenderHint(QPainter::Antialiasing);

    QColor c;
    for (int pass = 0; pass < 2; ++pass)
    {
        c = myForeground; c.setAlpha(32 + pass*48); p.setBrush(c);
        c = myBackground; c.setAlpha(c.alpha()/(6-4*pass)); p.setPen(c);
        for (int i = 0; i < 4; ++i)
        {
            p.resetTransform();
            p.translate(2 + i*d,2);
            for (int j = 0; j < 4; ++j)
            {
                p.translate(0,d);
                if (matrix[i][j] == pass && (i || j >1) && (j || i!=2))
                    p.drawPath(dot);
            }
        }
    }
    p.end();

    return img;
}

QImage *
BeClock::mayaImage( const QDateTime &dt )
{
    QPainterPath label;
    // HOUR ======================================================================
    int y = 60; int t = dt.time().hour() % 12;
    if (!t)
        label.addRoundedRect(10,10,10,65, 3,6);
    else
    {
        while (t > 4)
            { label.addRoundedRect(0,y,30,20, 4,6); y -= 30; t -= 5; }
        while (t > 1)
            { label.addRoundedRect(0,y,14,20, 4,6); label.addRoundedRect(16,y,14,20, 4,6); y -= 30; t -= 2; }
        if (t)
            label.addRoundedRect(0,y,14,20, 4,6);
    }

    // MINUTE DEC ==================================================================
    y = 60; t = dt.time().minute()/10;
    if (!t)
        label.addRoundedRect(45,10,10,65, 3,6);
    else
    {
        while (t > 1)
            { label.addRoundedRect(35,y,30,20, 4,6); y -= 30; t -= 2; }
        if (t)
            label.addRoundedRect(35,y,14,20, 4,6);
    }

    // MINUTE DIG ================================================================
    y = 60; t = dt.time().minute() % 10;
    if (!t)
        label.addRoundedRect(80,10,10,65, 3,6);
    else
    {
        while (t > 2)
            { label.addRoundedRect(70,y,30,20, 4,6); y -= 30; t -= 3; }
        while (t > 1)
            { label.addRoundedRect(70,y,14,20, 4,6); label.addRoundedRect(86,y,14,20, 4,6); y -= 30; t -= 2; }
        if (t)
            label.addRoundedRect(70,y,14,20, 4,6);
    }

    QImage *img = new QImage(myGeometry.size(), QImage::Format_ARGB32);
    img->fill( Qt::transparent );

    QPainter p( img );
    p.setRenderHint(QPainter::Antialiasing);
    QRect r(img->rect()); r.adjust(2,2,-2,-2);
    r.adjust(0, r.height()/6, 0, -r.height()/6);
    p.translate(r.topLeft());
    p.scale( (r.width()-4)/100.0, (r.height()-4)/100.0 );
    QColor c = myForeground; c.setAlpha(72); p.setBrush(c);
    c = myBackground; c.setAlpha(c.alpha()/2); p.setPen(c);
    p.drawPath(label);
    p.end();

    return img;
}

QImage *
BeClock::ringImage( const QDateTime &dt )
{
    const bool showDigits = myCountDown || iShowDigits;

    QImage *img = new QImage(myGeometry.size(), QImage::Format_ARGB32);
    img->fill( Qt::transparent );
    QPainter p( img );

    QRect rect = img->rect();
    const int d = qMax( 2, rect.width()/24 );
    int d2 = 2*d;

    if ( showDigits )
    {
        int d3 = 9;
        if (iShowRings) d3 = 2*(d2+d);
        rect.adjust( d3, d3, -d3, -d3 );
        if (rect.width() > 32)
        {
            QImage *digits = digitImage( dt, rect );
            p.drawImage( 0,0, *digits );
            delete digits;
        }
        rect = img->rect();
    }

    if ( iShowRings )
    {
        int alphaFactor = 1;
        QColor c = myForeground;
        QColor bc = myBackground;
        const int alpha = c.alpha();

        p.setRenderHint( QPainter::Antialiasing );
        // minutes
        int m = dt.time().minute();
        if (!m) { m = 60; alphaFactor = 2; }
        rect.adjust(d,d,-d,-d);
        c.setAlpha( alpha/(3*alphaFactor) );
        bc.setAlpha( 72/alphaFactor );
        drawOutlinedArc( p, rect, 90*16, m*-96, d, c, bc ); // 96 = 360*16/60

        // hours
        alphaFactor = 1;
        int h = dt.time().hour();
        if ( h > 12 )
            h -= 12;
        else if (!h)
            { h = 12; alphaFactor = 2; }

        rect.adjust(d2,d2,-d2,-d2);
        c.setAlpha( alpha*2/(3*alphaFactor) );
        bc.setAlpha( 128/alphaFactor );
        drawOutlinedArc( p, rect, 90*16, -480*h, d, c, bc ); // 480 = 360*16/12
        if (myHomeTimeDiff && !myCountDown) {
            bc.setAlpha( c.alpha() );
            drawOutlinedArc( p, rect, 90*16 -480*h + 8*(myHomeTimeDiff - (m%60)) - 3*8, 3*16, d, c, bc ); // 480 = 360*16/12, 8 = 360*16/12*60
        }

        if ( iShowTheDate && !showDigits )
        {
            const int day = dt.date().day();
            const int month = dt.date().month();

            // day
            rect.adjust(d2,d2,-d2,-d2);
            c.setAlpha( alpha*2/3 );
            bc.setAlpha( 128 );
            drawOutlinedArc( p, rect, 90*16, day*-360*16/dt.date().daysInMonth(), d, c, bc );

            // month
            rect.adjust(d2,d2,-d2,-d2);
            c.setAlpha( alpha/3 );
            bc.setAlpha( 72 );
            drawOutlinedArc( p, rect, 90*16, month*-480, d, c, bc ); // 480 = 360*16/12
        }
    }

    p.end();
    return img;
}

QImage *
BeClock::clockImage()
{
    QDateTime dt;

    if ( myCountDown )
    {
        int m = myCountDown/60;
        int s = 0;
        if ( m > 23 )
        {
            s = m%60;
            m = m/60;
        }
        else
            s = myCountDown - 60*m;
        dt.setTime( QTime(m, s) );
    }
    else
        dt = QDateTime::currentDateTime();

    switch (myType)
    {
        case Rings:
        default:
            return ringImage(dt);
        case Binary:
            return binaryImage(dt);
        case Maya:
            return mayaImage(dt);
    }
}

void
BeClock::updateBuffer()
{
    QImage *img = clockImage();

    if ( effects->compositingType() == OpenGLCompositing )
    {
        if ( myBufferGL[0] )
            myBufferGL[0]->load( *img );
        else
            myBufferGL[0] = new GLTexture( *img );
    }

    if ( effects->compositingType() == XRenderCompositing )
    {
        delete myBufferXR[0];
        myBufferXR[0] = new QPixmap( QPixmap::fromImage( *img ) );
#if 0
        // NOTICE This does NOT work, as Qt stumbles on the antialiasing (apparently paintengine
        // switching even more broken :-(
        // also it appears that calling antialiasing seems to force the raster engine and then performs
        // a conversion??? - that's dull :-(
        myBufferXR->fill( Qt::transparent );
        QPainter p( myBufferXR );
        drawClock( p, myBufferXR->rect() );
        p.end();
#endif
    }
    delete img;
}

// #include "moc_beclock.cpp"
